-- TODO: Add the possibility to ignore alts sync

-- Creating AddOn
Crafts = LibStub("AceAddon-3.0"):NewAddon("Crafts", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceComm-3.0")

-- AddOn config
Crafts.debug = false

-- Setting used local variables
local L = LibStub("AceLocale-3.0"):GetLocale(Crafts.name, false)
local B = LibStub("LibBabble-Inventory-3.0"):GetUnstrictLookupTable()
local BR = LibStub("LibBabble-Inventory-3.0"):GetReverseLookupTable()
local TIME_OF_FIRST_SYNC = 300 -- 5mn
local TIME_BETWEEN_SYNC = 3600 -- 1h
local PLAYER_EXPIRY_TIME = 1209600 -- 2 weeks
local MAX_PLAYERS_PER_CRAFT = 12 -- Maximum number of players registered per craft
local SYNC_END = "SYNC_END:"
local EMPTY_SYNC_END = "EMPTY_SYNC_END:"
local Crafts, self	= Crafts, Crafts
local CraftsTooltip -- Hidden tooltip to get item info
local CraftsHelpFrame = nil -- Help frame, not created by default
local CraftsMWCraftsNumber = 9 -- Number of crafts displayed in the main window
local CraftsSelectedLineNumber = -1 -- Currently selected craft line number
local CraftsSelectedLineWidget = nil -- Currently selected craft line widget
local CraftsHighlightedLineWidget = nil -- Currently highlighted craft line widget
local CraftsSearchTable -- Table containing results of last search
local CraftsSyncTimer = nil -- Timer launching the sync method
local CraftsIsSyncing = false -- If the addon is currently syncing
local CraftsInitialization = nil -- Time of addon initialization (login or /reloadui)
local CraftsLimitedMode = false
local CraftsTradeskillsIgnored = {
	-- As foundable in Libs/LibBabble-Inventory-3.0/LibBabble-Inventory-3.0.lua
	"Cooking",
	"First Aid"
}
local CraftsSkillLevelDefaultLimitation = 430 -- Don't synchronize characters with skill < to this value
local CraftsSkillLevelLimitations = {
	-- Specific limitation values for some crafts
	-- 
	-- Warning: level specified are
	-- *included* in synchronization
	Alchemy = 430,
	Blacksmithing = 430,
	Enchanting = 430,
	Engineering = 430,
	Jewelcrafting = 430,
	Leatherworking = 430,
	Tailoring = 430
}
local CraftsDisplayedMessages = {}

-- Returns the localization variable
-- used by the addon
function Crafts:L()
	return L
end

-- Used to reset saved variables
function Crafts:ResetSavedVariables()
	CraftsDB = {
		crafts = {},
		firstrun = true,
		icon = true,
		lastsync = nil,
		locale = GetLocale(),
		minimap = {}
	}
end

-- When the addon first starts
function Crafts:OnInitialize()
	CraftsInitialization = time()

	-- Initialize saved variables
	if not CraftsDB then
		self:ResetSavedVariables()
	end
	
	-- Manage multi-language clients
	if CraftsDB.locale ~= GetLocale() then
		CraftsLimitedMode = true
	end
	
	-- Creating tooltip
	CraftsTooltip = CreateFrame("GameTooltip", "CraftsGameTooltip", UIParent, "GameTooltipTemplate")
	
	-- Adding slash commands
	local options = {
		type = "group",
		args = {
			debug = {
				type = "toggle",
				name = L.debug,
				desc = L.debug_desc,
				get = function(_)
					return Crafts.debug
				end,
				set = function(_, value)
					if CraftsLimitedMode then
						self:Print(L.nodebugmode)
						return
					end
					
					Crafts.debug = value
					if Crafts.debug then
						self:Print("|c"..self:MakeTextHexColor(1.0,0.1,0.1,1.0)..L.debug_on.."|r")
						self:Print("|c"..self:MakeTextHexColor(1.0,0.5,0.1,1.0)..L.debug_warn.."|r")
					else
						self:Print("|c"..self:MakeTextHexColor(0.1,1.0,0.1,1.0)..L.debug_off.."|r")
					end
				end,
				hidden = true
			},
			forcesync = {
				type = "execute",
				name = L.forcesync,
				desc = L.forcesync_desc,
				func = function()
					self:ForceSync()
				end,
				hidden = true
			},
			help = {
				type = "execute",
				name = L.help,
				desc = L.help_desc,
				func = function()
					self:ShowHelp()
				end
			},
			minimap = {
				type = "toggle",
				name = L.minimap,
				desc = L.minimap_desc,
				get = function(_)
					return CraftsDB.icon
				end,
				set = function(_, value)
					CraftsDB.icon = value
					if CraftsDB.icon then
						self:ShowMinimapIcon()
					else
						self:HideMinimapIcon()
					end
				end
			},
			show = {
				type = "execute",
				name = L.show,
				desc = L.show_desc,
				func = function()
					self:ShowMW()
				end
			},
			timeleft = {
				type = "execute",
				name = L.timeleft,
				desc = L.timeleft_desc,
				func = function()
					self:PrintTimeLeftBeforeSync()
				end
			},
			version = {
				type = "execute",
				name = L.version,
				desc = L.version_desc,
				func = function()
					self:Print(Crafts.name.." 0.7.3-beta")
				end
			}
		}
	}
	LibStub("AceConfig-3.0"):RegisterOptionsTable(Crafts.name, options, "crafts")
	
	-- Displaying minimap icon only if needed
	if CraftsDB.icon then
		self:ShowMinimapIcon()
	end
	
	-- Make main window closeable with escape
	tinsert(UISpecialFrames, "CraftsFrame")
	
	-- Schedule first update
	-- Note: do not move this
	if not CraftsLimitedMode then
		self:ScheduleAutoSync()
	end
end

-- When the addon is enabled
-- or when user reloads UI
function Crafts:OnEnable()
	-- Title of the addon
	local title = Crafts.name

	-- Register first run event trigger only if
	-- it is the actual first run
	if CraftsDB.firstrun then
		self:RegisterEvent("PLAYER_ENTERING_WORLD")
	end

	-- If limited mode, don't update player's craft
	-- and don't sync
	if not CraftsLimitedMode then
		self:RegisterEvent("TRADE_SKILL_SHOW")
		self:RegisterComm(Crafts.name)
	else
		title = title.." (|c"..self:MakeTextHexColor(1.0,0.1,0.1,1.0)..L.limited.."|r)"
		self:Print("|c"..self:MakeTextHexColor(1.0,0.1,0.1,1.0)..L.enablinglimitedmode.."|r")
	end
	
	-- For online/offline players automatic update
	self:RegisterEvent("GUILD_ROSTER_UPDATE")

	-- First update of search table
	self:UpdateSearchTable()
	
	-- Slider initialization
	getglobal("CraftsSlider"):SetScript("OnValueChanged", function()
		getglobal(this:GetName()..'Text'):SetText(this:GetValue());
		self:UpdateMW()
	end)
	
	-- Players frame initialization
	getglobal("CraftsPlayersFrameTitle"):SetText('<html><body><p align="center">'..L.listofcrafters..'</p></body></html>')
	getglobal("CraftsAddonTitle"):SetText('<html><body><p align="center">'..title..'</p></body></html>')
	
	-- First action: update slider
	self:UpdateSlider()
end

-- UI events of first run should be managed here,
-- otherwise UI elements may not be displayed
-- (<frame>:Show(), ...)
function Crafts:PLAYER_ENTERING_WORLD()
	-- If first run, display help
	if CraftsDB.firstrun then
		self:ShowHelp()
		CraftsDB.firstrun = false
	end
end

-- Updates the slider of the addon, modifying the
-- max value
function Crafts:UpdateSlider()
	-- Remove current selection and hide crafters
	-- window since the content of the slider will change
	self:ClearSelectedLine()
	
	-- Update the slider
	local slider = getglobal("CraftsSlider")
	local max
	
	_, max = slider:GetMinMaxValues() 
	
	-- Show/hide the slider, update its min/max values
	if self:GetSearchedCraftsCount()<1 then
		-- No items in the list. Hide the slider.
		if slider:IsVisible() then
			slider:Hide()
		end
		
		self:UpdateMW()
	else
		-- Update only if necessary (max value has changed)
		
		slider:SetMinMaxValues(1,self:GetSearchedCraftsCount())
		getglobal("CraftsSliderHigh"):SetText(self:GetSearchedCraftsCount())
		
		if not slider:IsVisible() then
			slider:Show()
		end
	
		self:UpdateMW()
	end
end

-- Shows the help window
function Crafts:ShowHelp()
	if not CraftsHelpFrame then
		CraftsHelpFrame = CreateFrame("Frame", "CraftsHelp", UIParent, "CraftsHelpV")
		local string = ""
		string = string.."<html>"
		string = string.."<body>"
		string = string.."<h1 align='center'>"..L.welcome.."</h1>"
		string = string.."<p align='center'>"..L.purpose.."</p>"
		string = string.."<br/>"
		string = string.."<p>- \""..L.q1.."\"<br/>"
		string = string.."- \""..L.q2.."\"<br/>"
		string = string.."- \""..L.q3.."\"</p>"
		string = string.."<br/>"
		string = string.."<p align='center'>"..L.syncpreview.."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.howtoshare.."</h2>"
		string = string.."<p>"..L.howtoshare_desc.."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.howtosearch.."</h2>"
		string = string.."<p>"..format(L.howtosearch_desc, "|c"..self:MakeTextHexColor(0.8,0.8,0.1,1.0).."/crafts show|r").."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.whensyncis.."</h2>"
		string = string.."<p>"..L.whensyncis_desc.."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.whennextsyncis.."</h2>"
		string = string.."<p>"..format(L.whennextsyncis_desc, "|c"..self:MakeTextHexColor(0.8,0.8,0.1,1.0).."/crafts timeleft|r").."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.removeminimapicon.."</h2>"
		string = string.."<p>"..format(L.removeminimapicon_desc, "|c"..self:MakeTextHexColor(0.8,0.8,0.1,1.0).."/crafts minimap|r").."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.notallsynced.."</h2>"
		string = string.."<p>"..format(L.notallsynced_desc, "|c"..self:MakeTextHexColor(0.1,1.0,0.1,1.0).."Favorites.lua|r").."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.whatislimitedmode.."</h2>"
		string = string.."<p>"..L.whatislimitedmode_desc.."</p>"
		string = string.."<br/>"
		string = string.."<p>"..format(L.whatislimitedmode_desc2, "|c"..self:MakeTextHexColor(0.1,1.0,0.1,1.0).."WTF/Account/&lt;Login&gt;/SavedVariables/Crafts.lua|r").."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.stillhavequestion.."</h2>"
		string = string.."<p>"..L.stillhavequestion_desc.."</p>"
		string = string.."<br/>"
		string = string.."<h2 align='center'>"..L.donations.."</h2>"
		string = string.."<p>"..format(L.donations_desc, "|c"..self:MakeTextHexColor(0.1,1.0,0.1,1.0).."donations.txt|r").."</p>"
		string = string.."</body>"
		string = string.."</html>"
		
		local html = getglobal("CraftsHelpScrollFrameHTML")
		local font = html:GetFont()
		html:SetFont("h1", font, 28)
		html:SetTextColor("h1", 1.0, 1.0, 0.1, 1.0)
		html:SetFont("h2", font, 20)
		html:SetTextColor("h2", 1.0, 0.6, 0.1, 1.0)
		html:SetFont("p", font, 15)
		
		-- Must be done at the end
		html:SetText(string)
		
		-- Make help window closeable with escape
		tinsert(UISpecialFrames, "CraftsHelp")
	end
	CraftsHelpFrame:Show()
end

--Returns a pointer to the main window
function Crafts:MW()
	return getglobal("CraftsFrame")
end

-- Shows the main window
function Crafts:ShowMW()
	if not self:MW():IsVisible() then
		self:MW():Show()
	end
end

-- Hides the main window
function Crafts:HideMW()
	if self:MW():IsVisible() then
		self:MW():Hide()
	end
end

-- Shows the main window if it is
-- hidden, hides it otherwise 
function Crafts:ToggleMW()
	if self:MW():IsVisible() then
		self:HideMW()
	else
		self:ShowMW()
	end
end

-- Updates the crafts displayed in
-- the main window
function Crafts:UpdateMW()
	local craft
	local craftline
	
	-- Remove any green backdrop that
	-- was previousy selected
	if CraftsSelectedLineWidget ~= nil then
		self:RemoveBackdrop(CraftsSelectedLineWidget)
	end
	
	for i=0,CraftsMWCraftsNumber-1 do
		craftline = getglobal("CraftLine"..(i+1))
		craft = CraftsSearchTable[getglobal("CraftsSlider"):GetValue()+i]
		
		if craft ~= nil then
			-- If current line is the selected one
			if getglobal("CraftsSlider"):GetValue()+i == CraftsSelectedLineNumber then
				-- Update the reference to the selected widget
				CraftsSelectedLineWidget = craftline
				-- Display a nice backdrop on the widget
				self:SetGreenBackdrop(CraftsSelectedLineWidget)
			end
		
			-- Craft line tooltip
			craftline:SetScript("OnMouseDown", function()
				if arg1 ~= "LeftButton" then
					return
				end
				
				-- Clear focus from search box
				getglobal("CraftsEditBoxSearch"):ClearFocus()
			
				-- Remove previously set backdrop if applicable
				if CraftsSelectedLineWidget ~= nil then
					self:RemoveBackdrop(CraftsSelectedLineWidget)
				end
				
				-- Set special backdrop on selection &
				-- save which line is selected
				self:SetGreenBackdrop(this)
				CraftsSelectedLineWidget = this
				CraftsSelectedLineNumber = getglobal("CraftsSlider"):GetValue() + string.sub(this:GetName(), string.len(this:GetName())) - 1
				
				-- Set craft name in crafters window
				getglobal("CraftsPlayersCraftName"):SetText(
					'<html><body><p align="center">'
						..GetSpellLink(CraftsSearchTable[getglobal("CraftsSlider"):GetValue()+i].hyperlink)..
					'</p></body></html>'
				)
			
				-- Set online & offline crafters in
				-- crafters window
				self:UpdateCraftOnlineAndOfflinePlayers()
				
				-- Calling guild roster update if available
				GuildRoster()
				
				-- Display crafters window
				local playersframe  = getglobal("CraftsPlayersFrame")
				if not playersframe:IsVisible() then
					playersframe:Show()
				end
			end)
			craftline:SetScript("OnEnter", function()
				CraftsHighlightedLineWidget = this
				if CraftsSelectedLineWidget ~= this then
					self:SetOrangeBackdrop(this)
				end
				self:ShowTooltip(
					this,
					'ANCHOR_LEFT',
					-- Cannot use 'craft' here, need the real reference
					"enchant:"..CraftsSearchTable[getglobal("CraftsSlider"):GetValue()+i].hyperlink
				)
			end)
			craftline:SetScript("OnLeave", function()
				CraftsHighlightedLineWidget = nil
				if CraftsSelectedLineWidget ~= this then
					self:RemoveBackdrop(this)
				end
				self:UnsetTooltip()
			end)
		
			-- Icon
			local icon =  getglobal("CraftLine"..(i+1).."Icon")
			icon:SetBackdrop({
				bgFile = "Interface/Icons/"..craft.texture, 
				edgeFile = nil, tile = false, tileSize = 16, edgeSize = 0, 
  				insets = { left = 0, right = 0, top = 0, bottom = 0 }
			})
			
			-- Title
			getglobal("CraftLine"..(i+1).."Title"):SetText(self:TruncateString(craft.name, 50))
			
			-- Player names
			local playernames = self:ImplodeElementsValues(", ", craft.players, "name")
			if playernames ~= "" then
				playernames = self:TruncateString(playernames, 60)
			else
				playernames = "--"
			end
			getglobal("CraftLine"..(i+1).."PlayerNames"):SetText(playernames)
			
			if not craftline:IsVisible() then
				craftline:Show()
			end
		else
			if craftline:IsVisible() then
				craftline:Hide()
			end
		end
	end
end

-- Updates the names of players of the current selected
-- craft in the crafters window, seperating them in an
-- online and an offline category
function Crafts:UpdateCraftOnlineAndOfflinePlayers()
	local connected, offline = self:SplitPlayersList(CraftsSearchTable[CraftsSelectedLineNumber].players)
	getglobal("CraftsPlayersConnected"):SetText(connected)
	getglobal("CraftsPlayersOffline"):SetText(offline)
end

-- Clears the selection on the currently
-- selected craft line.
function Crafts:ClearSelectedLine()
	if CraftsSelectedLineWidget ~= nil then
		self:RemoveBackdrop(CraftsSelectedLineWidget)
		CraftsSelectedLineWidget = nil
		CraftsSelectedLineNumber = -1
	end
end

-- Returns the line currently
-- hovered by the mouse
function Crafts:GetHoveredLine()
	return CraftsHighlightedLineWidget
end

-- Makes sure that the given string is not longer
-- than the given length, or else truncates and adds
-- "..." to it
function Crafts:TruncateString(string, maxsize)
	if string.len(string)>maxsize then
		string = string.sub(string, 1, maxsize-6).." (...)"
	end
	return string
end

-- Join table values with a glue string.
function Crafts:ImplodeValues(glue, table)
	local str = ""
	for _,v in pairs(table) do
		str = str..v..glue
	end
	str = string.sub(str, 1, -string.len(glue)-1)
	return str
end

-- Join the given table elements values indexed
-- by the given key with a glue string.
function Crafts:ImplodeElementsValues(glue, table, key)
	local str = ""
	
	for _,v in pairs(table) do
		str = str..v[key]..glue
	end
	str = string.sub(str, 0, -string.len(glue)-1)
	return str
end

-- Splits the given list of players into
-- two tables: connected and offline players.
--
-- The table containing the connected players
-- is enhanced with color classes and links to
-- easily send whispers to them.
function Crafts:SplitPlayersList(players)
	local connected = {}
	local offline = {}
	local r, g, b, a
	local online, englishclass
	
	for _,v in pairs(players) do
		online, englishclass = self:IsGuildMemberConnected(v.name)
		r, g, b, a = RAID_CLASS_COLORS[englishclass].r, RAID_CLASS_COLORS[englishclass].g, RAID_CLASS_COLORS[englishclass].b, 1.0
		
		if online then
			table.insert(connected, "|c"..self:MakeTextHexColor(r, g, b, a).."|Hplayer:"..v.name.."|h["..v.name.."]|h|r")
		elseif englishclass ~= nil then
			table.insert(offline, "|c"..self:MakeTextHexColor(r, g, b, a)..v.name.."|r")
		else
			table.insert(offline, v.name)
		end
	end
	
	return self:ImplodeValues(", ", connected), self:ImplodeValues(", ", offline)
end

-- Checks if the given guild member is connected
-- to the game. This method also returns the
-- english class name of the player, whether
-- it is connected or not.
--
-- If the player is not a member of the guild of
-- the player, false is returned as well as a
-- nil english class name.
function Crafts:IsGuildMemberConnected(name)
	local guildmembername, online, englishclass
	for i=1,GetNumGuildMembers(true) do
		guildmembername, _, _, _, _, _, _, _, online, _, englishclass = GetGuildRosterInfo(i)
		if guildmembername == name then
			return online, englishclass
		end
	end
	return false, nil
end

-- Prints the message identified by the given
-- format string only if it has not already been
-- printed by using this method.
--
-- This method can take any number of arguments
-- after the format string as the format()
-- function would do.
--
-- The "not/already displayed" status of each
-- message is reset on logout or /console reloadui
-- call.
function Crafts:PrintOnce(formatstring, ...)
	if not CraftsDisplayedMessages[formatstring] then
		if ... == nil then
			self:Print(formatstring)
		else
			self:Print(format(formatstring, unpack({...})))
		end
		CraftsDisplayedMessages[formatstring] = true
	end
end

-- This method prints a message to the user the same
-- way as Crafts:PrintOnce() does, but formats the
-- format string prior to checking if the message
-- has been displayed or not.
--
-- You may by this way use the same format string for
-- different messages but be sure that the message
-- will be displayed once for each pack of parameters. 
function Crafts:PrintOnceWithFormatting(formatstring, ...)
	if ... == nil then
		self:PrintOnce(formatstring)
		return
	end

	local msg = format(formatstring, unpack({...}))

	if not CraftsDisplayedMessages[msg] then
		self:Print(msg)
		CraftsDisplayedMessages[msg] = true
	end
end

-- Makes a color hexadecimal string from the
-- given red, green, blue and alpha values.
--
-- It may then be used in a "|c<hex><text>|r"
-- string.
function Crafts:MakeTextHexColor(r, g, b, a)
	local ir, ig, ib = string.format("%02x", r*255), string.format("%02x", g*255), string.format("%02x", b*255)
	local ia
	if a then
		ia = string.format("%02x", a*255)
	else
		ia = "ff"
	end
	
	return (ia .. ir .. ig .. ib)
end

-- Builds a craft with the given properties and
-- and registers it.
--
-- This method returns false if the craft could
-- not be built (and thus, registered), true
-- otherwise.
function Crafts:BuildAndRegisterCraft(craftname, crafthyperlink, crafttype, crafttexture, playername)
	if craftname == nil
		or crafthyperlink == nil
		or crafttype == nil
		or crafttexture == nil
		or playername == nil then
		return false -- KO
	end
	
	self:RegisterCraft(self:BuildCraft(craftname, crafthyperlink, crafttype, crafttexture, playername))
	return true -- OK
end

-- Registers a craft in the internal list of crafts.
--
-- 1)
-- If the craft does not exist, it is created.
-- If it exists, nothing is modified.
-- 
-- 2)
-- If the players names contained in the given
-- craft are not associated to the internal list
-- craft, they are associated. Otherwise, only
-- their last update time is updated.
--
-- This method returns -1 if the craft did not
-- need any registration or update, 0 if it was
-- updated, and 1 if it was added to the internal
-- list.
function Crafts:RegisterCraft(craft)
	local result = -1 -- Default: nothing has been changed

	-- Registering craft
	local craftpos, craftadded = self:SortAdd(CraftsDB.crafts, craft, "name")

	if not craftadded then
		-- Craft not added, must add new players to players list
		
		local playerpos, playeradded
		for _,v in ipairs(craft.players) do
			playerpos, playeradded = self:SortAdd(CraftsDB.crafts[craftpos].players, v, "name")

			if not playeradded then
				-- Player not added, make sure 'update' key is up-to-date
				
				if v.update > CraftsDB.crafts[craftpos].players[playerpos].update then
					CraftsDB.crafts[craftpos].players[playerpos].update = v.update
				end
			elseif result ~= 0 then
				result = 0 -- Craft has been updated
			end
		end
	else
		result = 1 -- Craft has been added
	end
	
	-- Remove oldest player if number of players
	-- register exceeds MAX_PLAYERS_PER_CRAFT
	while #CraftsDB.crafts[craftpos].players > MAX_PLAYERS_PER_CRAFT do
		local oldest = nil
		local postoremove = -1
		for k,v in ipairs(CraftsDB.crafts[craftpos].players) do
			if oldest == nil or v.update < oldest then
				postoremove = k
				olest = v.update
			end
		end
		table.remove(CraftsDB.crafts[craftpos].players, postoremove)
		
		-- If result was "nothing changed", then change it to "updated"
		-- If result was "craft added", then leave it to this value
		-- Thus, the check on "== -1"
		if result == -1 then
			result = 0
		end
	end
	
	return result
end

-- Adds the given object to the given table,
-- in the order defined by the comparison between
-- the key of each table element and the
-- given object key that match the given key name.
--
-- For example, if you have a table of this type:
-- table = {
-- 	{
-- 		name = "abc",
--  	value = "def"
--	}, -- [1]
-- 	{
-- 		name = "test",
--  	value = "value"
-- 	} -- [2]
-- }
--
-- ...and the given object:
-- {
-- 	name = "other",
-- 	value = "val"
-- }
--
-- ...and you call this method with these parameters:
-- SortAdd(
-- 	table,
-- 	object,
-- 	"name"
-- )
--
-- ...this method would place the object in
-- second position since "other" is alphabetically
-- placed between "abc" and "test".
--
-- If the object is already in the table, then it
-- is ignored.
--
-- If the table is empty, the object is added to it.
--
-- This method returns the position at which the
-- object was added or found, and a boolean which
-- value is true if the given object was added,
-- or false if it was already in the given table.
--
-- NOTE: The given table must already be sorted.
function Crafts:SortAdd(t, object, keyname)
	if #t == 0 then
		table.insert(t, object)
		return 1, true
	else
		for i,v in ipairs(t) do
			-- Already added, just quit
			if object[keyname] == v[keyname] then
				return i, false
			end
			
			-- Craft name is located before
			-- current element of the list
			if object[keyname] < v[keyname] then
				table.insert(t, i, object)
				return i, true
			end
			
			-- We reached the end of the list,
			-- add it to the end
			if i == #t then
				table.insert(t, object)
				return i+1, true -- Added to #CraftsDB.crafts+1
			end
		end
	end
end

-- Builds a new craft element from the given
-- values, that should be stored in the internal
-- list of crafts used by the addon
function Crafts:BuildCraft(craftname, crafthyperlink, crafttype, crafttexture, playername)
	return {
		name = craftname,
		hyperlink = crafthyperlink,
		type = crafttype,
		texture = string.sub(crafttexture, 17),
		players = {
			{
				name = playername,
				update = time()
			}
		}
	}
end

-- When the guild roster is updated, the players
-- frame is updated with currently online/offline
-- characters if it is visible
function Crafts:GUILD_ROSTER_UPDATE()
	if CraftsPlayersFrame:IsVisible() then
		self:UpdateCraftOnlineAndOfflinePlayers()
	end
end

-- When the player opens the tradeskill window
-- 
-- This method parses the tradeskill window of
-- the player (and only him).
function Crafts:TRADE_SKILL_SHOW()
	if IsTradeSkillLinked() or IsTradeskillTrainer() then -- Linked or at trainer
		return
	end
	
	-- Retrieving information on tradeskill
	local tradeskillname, tradeskillcurrentlevel = GetTradeSkillLine()
	
	-- Ignoring crafts to ignore ;)
	for _,v in ipairs(CraftsTradeskillsIgnored) do
		if tradeskillname == B[v] then
			self:PrintOnceWithFormatting(L.craftignored, tradeskillname)
			return
		end
	end
	
	-- Checking skill level
	-- If too low, sync only favorites
	local favoritesonly = false
	local tradeskilllevellimitation = self:GetTradeskillLevelLimitation(tradeskillname)
	if tradeskillcurrentlevel < tradeskilllevellimitation then
		self:PrintOnce(L.skilltoolow, tradeskillname, tradeskilllevellimitation)
		favoritesonly = true
	end

	-- Number of crafts before update
	local craftsnumber = #CraftsDB.crafts

	-- Skills
	local skillname, skilltype
	local headerfound = false
	
	-- Items
	local itemlink, itemtexture, itemid
	
	-- Spells
	local spellcolor, spellid, spelltexture
	
	-- Iterator
	local i
	
	-- Iterate over all crafts
	for i=1, GetNumTradeSkills() do
		skillname, skilltype = GetTradeSkillInfo(i)
		
		if skilltype == "header" and skillname ~= nil then
			headerfound = true
		else
			if not headerfound then
				self:Print(L.tradeskillanalysefailed)
				return
			end
			
			-- Get the recipe link
			recipelink = GetTradeSkillRecipeLink(i)
			recipename, recipehyperlink = self:RecipeLinkDecode(recipelink)
			
			-- Sync stuff to sync
			-- * Force the sync of favorites
			-- * Sync all non trivial crafts if "favorites only" mode is not enabled (due to low skill level, ...)
			if (not favoritesonly and skilltype ~= "trivial") or self:IsRecipeFavorite(tradeskillname, recipehyperlink) then
				-- Get the item link of the craft
				itemlink = GetTradeSkillItemLink(i)
				
				if string.find(itemlink, "|Hitem:") then -- Item
					_, _, _, _, _, _, _, _, _, itemtexture = GetItemInfo(itemlink)
					
					if not self:BuildAndRegisterCraft(recipename, recipehyperlink, "i", itemtexture, UnitName("player")) then
						self:Print(format(L.reagentanalysefailed, GetTradeSkillLine()))
						self:ParseTradeSkill()
						return
					end
				else -- Spell
					spelltexture = "Interface\\Icons\\Spell_Holy_GreaterHeal"
					
					if not self:BuildAndRegisterCraft(recipename, recipehyperlink, "s", spelltexture, UnitName("player")) then
						self:Print(format(L.reagentanalysefailed, GetTradeSkillLine()))
						self:ParseTradeSkill()
						return
					end
				end
			end
		end
	end
	
	if #CraftsDB.crafts ~= craftsnumber then
		-- New crafts!
		self:UpdateSearchTable(getglobal("CraftsEditBoxSearch"):GetText())
		self:UpdateSlider()
	else
		self:UpdateMW() -- Not that long
	end
end

-- Checks if the given recipe hyperlink is considered
-- as a favorite for the given tradeskill. If either
-- the tradeskill or the recipe is unknown, false is
-- always returned.
function Crafts:IsRecipeFavorite(tradeskillname, recipehyperlink)
	if BR[tradeskillname] == nil or Crafts.favorites[BR[tradeskillname]] == nil then
		return false
	end
	
	for _,v in ipairs(Crafts.favorites[BR[tradeskillname]]) do
		if v == tonumber(recipehyperlink) then
			return true
		end
	end
	return false
end

-- Returns the level limitation for the given
-- tradeskill. If a limitation for the given
-- tradeskill is not found, a default value
-- is returned.
function Crafts:GetTradeskillLevelLimitation(tradeskillname)
	if CraftsSkillLevelLimitations[tradeskillname] ~= nil then
		return CraftsSkillLevelLimitations[tradeskillname]
	else
		return CraftsSkillLevelDefaultLimitation
	end
end

-- Decodes an enchant link and returns
-- the id and name of the enchant
function Crafts:EnchantLinkDecode(enchantlink)
	local enchantid, enchantname
	for enchantid, enchantname in string.gmatch(enchantlink, "|c%x+|Henchant:(%d+)|h%[(.+)%]|h|r") do
		return enchantid, enchantname
	end
end

-- Decodes a recipe link and returns
-- the name and hyperlink of the enchant
function Crafts:RecipeLinkDecode(recipelink)
	local recipeid, recipename = self:EnchantLinkDecode(recipelink)
	return recipename, recipeid
end

-- Parses the current opened tradeskill
-- in order to force an update of the
-- client's cache
function Crafts:ParseTradeSkill()
	local skilltype
	
	-- Iterators
	local i, reagent
	
	for i=1, GetNumTradeSkills() do
		_, skilltype, _, _ = GetTradeSkillInfo(i)
		
		if skilltype ~= "header" then
			CraftsTooltip:SetTradeSkillItem(i)
			
			for reagent=1, GetTradeSkillNumReagents(i) do
				CraftsTooltip:SetTradeSkillItem(i, reagent)
			end
		end
	end
end

-- Returns the number of known crafts in the database
function Crafts:GetKnownCraftsCount()
	return #CraftsDB.crafts
end

-- Returns the number of crafts found during the
-- last search
function Crafts:GetSearchedCraftsCount()
	return #CraftsSearchTable
end

-- Adds a nice orange backdrop to the given widget
function Crafts:SetOrangeBackdrop(widget)
	self:SetBackdrop(widget, "Interface/BUTTONS/YELLOWORANGE64")
end

-- Adds a nice blue backdrop to the given widget
function Crafts:SetBlueBackdrop(widget)
	self:SetBackdrop(widget, "Interface/BUTTONS/BLUEGRAD64")
end

-- Adds a nice green backdrop to the given widget
function Crafts:SetGreenBackdrop(widget)
	self:SetBackdrop(widget, "Interface/BUTTONS/GREENGRAD64")
end

-- Adds the given backdrop to the given widget
function Crafts:SetBackdrop(widget, file)
	widget:SetBackdrop( {
		bgFile = file,
		edgeFile = nil, tile = false, tileSize = 16, edgeSize = 0,
		insets = { left = 0, right = 0, top = 0, bottom = 0} 
	})
end

-- Removes the backdrop of the given widget
function Crafts:RemoveBackdrop(widget)
	widget:SetBackdrop(nil)
end

-- Shows the game tooltip on the given
-- widget, at the given anchor and with
-- the given hyperlink
function Crafts:ShowTooltip(widget, anchor, hyperlink)
	GameTooltip:SetOwner(widget, anchor)
	GameTooltip:SetHyperlink(hyperlink)
	GameTooltip:Show()
end

-- Shows the game tooltip on the given
-- widget, at the given anchor and with
-- the given single-line text
function Crafts:ShowTooltipMessage(widget, anchor, text)
	GameTooltip:SetOwner(widget, anchor)
	GameTooltip:ClearLines()
	GameTooltip:AddLine(text)
	GameTooltip:Show()
end

-- Hides the game tooltip, wherever it is
function Crafts:UnsetTooltip()
	GameTooltip:Hide()
end

-- Removes the category name of the given craft name
function Crafts:RemoveCraftCategory(craft)
	local firstcolon = string.find(craft, ':')
	return string.sub(craft, firstcolon+2)
end

-- Updates the internal search table, containing
-- the results of the last search
function Crafts:UpdateSearchTable(search)
	if search == nil then
		CraftsSearchTable = CraftsDB.crafts
		return
	end
	
	-- Reseting contents of search table
	CraftsSearchTable = {}
	
	-- Avoid lowering search at each iteration
	search = string.lower(search)
	
	local v
	for _,v in ipairs(CraftsDB.crafts) do
		if string.find(string.lower(v.name), search) ~= nil then
			table.insert(CraftsSearchTable, v)
		end
	end
end

-- Synchronization method
function Crafts:Sync()
	-- No syncrhonization in limited mode
	if CraftsLimitedMode then
		self:Print("|c"..self:MakeTextHexColor(1.0,0.5,0.1,1.0)..L.nosyncinlimitedmode.."|r")
		return
	end

	-- Don't sync if already syncing
	if CraftsIsSyncing then
		self:Print(L.alreadysyncing)
		return
	end
	
	-- Check that we're not trying to sync too soon (...Executus)
	if self:TimeLeftBeforeSync() > 1 then
		self:Print(L.cannotlaunchsyncnow)
		self:PrintTimeLeftBeforeSync()
		return
	end
	
	-- Starting sync
	CraftsIsSyncing = true
	
	local syncendstring

	if #CraftsDB.crafts > 0 then
		local craft
		local playerdeleted = false
		
		for i=1,#CraftsDB.crafts do
			craft = CraftsDB.crafts[i]
			if craft ~= nil then
				-- Delete "too old" players
				for k,v in ipairs(craft.players) do
					if time() - v.update > PLAYER_EXPIRY_TIME then
						if not playerdeleted then
							playerdeleted = true
						end
						table.remove(craft.players, k)
					end
				end
				
				-- Update main window if at least one player
				-- was deleted for display coherence
				if playerdeleted then
					self:UpdateMW()
				end
				
				-- Now synchronize the whole thing
				for _,v in ipairs(self:Serialize(craft)) do
					-- Bulk: may send more than 1000 messages
					self:SendCommMessage(Crafts.name, v, "GUILD", nil, "BULK")
				end
			end
		end
		
		syncendstring = SYNC_END
	else
		syncendstring = EMPTY_SYNC_END
	end
	
	-- "GUILD" mandatory here, to make sure it is
	-- received as last synchronization message
	--
	-- Player name is appended to be able to identify
	-- that the message comes from this client
	self:SendCommMessage(Crafts.name, syncendstring..UnitName("player"), "GUILD", nil, "BULK")
end

-- Synchronization from other addons
function Crafts:OnCommReceived()
	-- arg2: Message
	-- arg4: Player name
	
	local playername = UnitName("player")
	
	-- If sync end message is detected and
	-- belongs to current player
	if arg2 == SYNC_END..playername or arg2 == EMPTY_SYNC_END..playername then
		if arg2 == SYNC_END..playername then
			-- Update done, timestamping now as last sync
			CraftsDB.lastsync = time()
		end
		
		-- Schedule another update
		self:ScheduleAutoSync()
		
		-- Syncing over
		CraftsIsSyncing = false
	end
	
	-- Any sync end message received exits this method
	if self:StartsWith(arg2, SYNC_END) or self:StartsWith(arg2, EMPTY_SYNC_END) then
		if Crafts.debug then
			self:Print(arg2)
		end
		return
	end
	
	if Crafts.debug then
		self:Print(arg4..">"..arg2)
	end
	
	if arg4 ~= UnitName("player") then
		local craft = self:Unserialize(arg2)
		
		-- Invalid craft, ignore that
		if craft == nil and Crafts.debug then
			self:Print("|c"..self:MakeTextHexColor(1.0,0.5,0.1,1.0)..IGNORED.."|r")
			return
		end
		
		-- Getting craft name
		CraftsGameTooltip:SetOwner(UIParent, "ANCHOR_NONE")
		CraftsGameTooltip:SetHyperlink("enchant:"..craft.hyperlink)
		craft.name = CraftsGameTooltipTextLeft1:GetText()
		
		local result = self:RegisterCraft(craft)
		if result == 1 then
			-- Craft added
			-- Update search table with current search
			self:UpdateSearchTable(getglobal("CraftsEditBoxSearch"):GetText())
			self:UpdateSlider() -- Calls self:UpdateMW()
		elseif result == 0 then
			-- Craft updated
			self:UpdateMW()
		end
	end
end

-- Returns true if the given string starts with
-- the given prefix, false otherwise.
function Crafts:StartsWith(string, prefix)
	return string.find(string, prefix) == 1
end

-- Serializes the given craft into one or several
-- strings for easy export on the chat system
function Crafts:Serialize(craft)
	local tosend = {}
	local msghead = craft.type.."#"..craft.hyperlink.."#"..craft.texture.."#"
	local msg = msghead
	
	for _,v in ipairs(craft.players) do
		if string.len(msg..v.name..":"..v.update) > 200 then
			table.insert(tosend, string.sub(msg, 1, string.len(msg)-1))
			msg = msghead..v.name..":"..v.update..";"
		else
			msg = msg..v.name..":"..v.update..";"
		end
	end
	table.insert(tosend, string.sub(msg, 1, string.len(msg)-1))
	
	return tosend
end

-- Unserializes the given string into a craft
-- for easy import from the chat system. This
-- method returns nil if string could not
-- be unserialized.
function Crafts:Unserialize(string)
	local type, hyperlink, texture, players = string.match(string, "([is])#(%d+)#([%a%d_]+)#(.+)")
	
	-- Checking parsed variables and returning
	-- nil if an error occurred
	if type == nil or hyperlink == nil or texture == nil or players == nil then
		return nil
	end
	
	local craft = {
		type = type,
		hyperlink = hyperlink,
		texture = texture,
		players = {}
	}
	for playername, playerupdate in string.gmatch(players, "([^:]+):([%d]+);?") do
		-- Invalid sync message, abort
		if playername == nil or playerupdate == nil then
			return nil
		end
		tinsert(craft.players, {
			name = playername,
			update = tonumber(playerupdate)
		})
	end
	
	return craft
end

-- Cancels any sync timer currently running
-- and determines and schedules a new sync
-- timer triggering the sync method as
-- soon as possible.
--
-- The sync method cannot be called when:
--  * less than TIME_OF_FIRST_SYNC seconds
--    have passed since the last addon
--    initialization
--  * less than TIME_BETWEEN_SYNC seconds
--    have passend since the last sync
function Crafts:ScheduleAutoSync()
	-- Sanity check
	self:CancelTimer(CraftsSyncTimer, true)

	local nextsync = 0
	if CraftsInitialization + TIME_OF_FIRST_SYNC > time() then
		-- We're still before time of first sync allowed
		nextsync = CraftsInitialization + TIME_OF_FIRST_SYNC - time() -- Next asap
	end
	
	if CraftsDB.lastsync ~= nil then
		if time() + nextsync < CraftsDB.lastsync + TIME_BETWEEN_SYNC then
			-- We have to push sync even later to avoid
			-- syncing too many times per gaming session
			nextsync = CraftsDB.lastsync + TIME_BETWEEN_SYNC - time() -- TIME_BETWEEN_SYNC after last sync
		end
	end
	
	-- Sanity check
	if nextsync == 0 then
		-- We do not have a last sync time but we are
		-- after TIME_OF_FIRST_SYNC. Schedule next sync
		-- TIME_OF_FIRST_SYNC later.
		nextsync = TIME_OF_FIRST_SYNC
	end
	
	CraftsSyncTimer = self:ScheduleTimer("Sync", nextsync)
end

-- Forces an immediate synchronization
function Crafts:ForceSync()
	-- Sanity check
	self:CancelTimer(CraftsSyncTimer, true)

	self:Print(L.forcingsyncnow)
	CraftsSyncTimer = self:ScheduleTimer("Sync", 0)
end

-- Returns the number of seconds
-- until the next update, or a
-- localized message if the addon
-- is syncing
function Crafts:TimeLeftBeforeSync()
	if CraftsIsSyncing then
		return L.syncing
	elseif CraftsLimitedMode then
		return L.syncdisabled
	else
		return self:TimeLeft(CraftsSyncTimer)
	end
end

-- Returns the number of seconds
-- rounded to the nearest integer
-- until the next update, or a
-- localized message if the addon
-- is syncing
function Crafts:RoundedTimeLeftBeforeSync()
	local timeleft = self:TimeLeftBeforeSync()
	
	if type(timeleft) == "number" then
		timeleft = format("%i", timeleft+0.5)
	end
	
	return timeleft
end

-- Prints the time left before next sync
-- to the chat window
function Crafts:PrintTimeLeftBeforeSync()
	self:Print(L.remainingtime..self:TimeLeftBeforeSync())
end